#include <math.h>
/*G. Conant added include for c++ compile*/
#include <stdio.h>

/* The following code for calculating chi-squared probabilities is
   derived from code in "Numerical Recipes in C" - Press, et. al. */

double chisq_prob(double df, double chisq)
{
    void gcf(double*,double,double,double*);
    void gser(double*,double,double,double*);
    double gamser,gammcf,gln;

    chisq /= 2.0; df /= 2.0;
    if (chisq < (df+1.0)) {
        gser(&gamser,df,chisq,&gln);
	return 1.0-gamser;
    } else {
        gcf(&gammcf,df,chisq,&gln);
	return gammcf;
    }
}

#define ITMAX 1000
#define EPS 3.0e-7
#define FPMIN 1.0e-30

void gcf(double *gammcf, double a, double x, double *gln) 
{ 
    double gammln(double); 
    int i; 
    double an,b,c,d,del,h;

    *gln=gammln(a);
    b=x+1.0-a;
    c=1.0/FPMIN;
    d=1.0/b;
    h=d;
    for (i=1;i<=ITMAX;i++) {
        an = -i*(i-a);
	b += 2.0;
	d=an*d+b;
	if (fabs(d) < FPMIN) d=FPMIN;
	c=b+an/c;
	if (fabs(c) < FPMIN) c=FPMIN;
	d=1.0/d;
	del=d*c;
	h *= del;
	if (fabs(del-1.0) < EPS) break;
    }
/*    if (i > ITMAX) nrerror("a too large, ITMAX too small in gcf"); */

    *gammcf=exp(-x+a*log(x)-(*gln))*h;
}

void gser(double *gamser, double a, double x, double *gln)
{
    double gammln(double);
    int n;
    double sum,del,ap;

    *gln=gammln(a);
    ap=a;
    del=sum=1.0/a;
    for (n=1;n<=ITMAX;n++) {
        ++ap;
	del *= x/ap;
	sum += del;
	if (fabs(del) < fabs(sum)*EPS) {
	    *gamser=sum*exp(-x+a*log(x)-(*gln));
	    return;
	}
    }
/*    nrerror("a too large, ITMAX too small in routine gser"); */
    return;
}

double gammln(double xx)
{
	double x,y,tmp,ser;
	static double cof[6]={76.18009172947146,-86.50532032941677,
		24.01409824083091,-1.231739572450155,
		0.1208650973866179e-2,-0.5395239384953e-5};
	int j;

	y=x=xx;
	tmp=x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser=1.000000000190015;
	for (j=0;j<=5;j++) ser += cof[j]/++y;
	return -tmp+log(2.5066282746310005*ser/x);
}

#define TINY 1.0e-30

void cntab1(int **nn, int ni, int nj, double *chisq, double *df, double *prob,
	    double *cramrv, double *ccc)
{
        double gammq(double a, double x);
        int nnj,nni,j,i,minij;
        double sum=0.0,expctd,sumi[100],sumj[100],temp;

	
        nni=ni;
        nnj=nj;
        for (i=0;i<ni;i++) {
                sumi[i]=0.0;
                for (j=0;j<nj;j++) {
                        sumi[i] += nn[i][j];
                        sum += nn[i][j];
                }
                if (sumi[i] == 0.0) --nni;
        }
        for (j=0;j<nj;j++) {
                sumj[j]=0.0;
                for (i=0;i<ni;i++) sumj[j] += nn[i][j];
                if (sumj[j] == 0.0) --nnj;
        }
        *df=nni*nnj-nni-nnj+1;
        *chisq=0.0;

	if ((*df) < 0.5) {
	  *prob=1.0;
	  *cramrv = *ccc = 0.0;
	} else {

	  for (i=0;i<ni;i++) {
                for (j=0;j<nj;j++) {
                        expctd=sumj[j]*sumi[i]/sum;
                        temp=nn[i][j]-expctd;
                        *chisq += temp*temp/(expctd+TINY);
                }
	  }
	  *prob=gammq(0.5*(*df),0.5*(*chisq));
	  minij = nni < nnj ? nni-1 : nnj-1;
	  *cramrv=sqrt(*chisq/(sum*minij));
	  *ccc=sqrt(*chisq/(*chisq+sum));
	}
}


double gammq(double a, double x)
{
        void gcf(double *gammcf, double a, double x, double *gln);
        void gser(double *gamser, double a, double x, double *gln);
        void nrerror(char error_text[]);
        double gamser,gammcf,gln;

        if (x < 0.0 || a <= 0.0) printf("Invalid arguments in routine gammq");

        if (x < (a+1.0)) {
                gser(&gamser,a,x,&gln);
                return 1.0-gamser;
        } else {
                gcf(&gammcf,a,x,&gln);
                return gammcf;
        }
}
